unit KA.Data.Ado.DBGrid;
interface
{*******************************************************************************
  exoSelectCheckBoxes and Ascending/Descending
  sort indicators is inspired by SMDBGrid from
  Shkolnik Mike
  FIDOnet: 2:463/106.14
  E-Mail:  mshkolnik@scalabium.com
*******************************************************************************}
{$R 'TAdoDBGrid.bmp'}
{$R 'TAdoDateTime.bmp'}
{$R 'TAdoHRollForm.bmp'}


uses
  System.Reflection,
  Borland.Vcl.Windows,
  Borland.Vcl.Messages,
  Borland.Vcl.SysUtils,
  Borland.Vcl.Classes,
  Borland.Vcl.Controls,
  Borland.Vcl.Graphics,
  Borland.Vcl.DB,
  Borland.Vcl.Grids,
  Borland.Vcl.DBGrids,
  Borland.Vcl.Menus,
  Borland.Vcl.Clipbrd,
  Borland.Vcl.ADOInt,
  Borland.Vcl.ADODB;

type
  ExOptionsEnum = (exoRowSizing, exoWrapTitles, exoLoadAllRecords, exoSelectCheckBoxes);
  ExOptionsSet  = Set of ExOptionsEnum;
  TAdoDBGrid = class(TDBGrid)
  private
    { Private declarations }
    FTopRect             : TRect;
    FClicked             : Boolean;
    FPopupMenu           : TPopupMenu;
    FMenuBMP             : TBitmap;
    FASCBMP              : TBitmap;
    FDESCBMP             : TBitmap;
    FMenuTitles          : TStringList;
    FOriginalTitles      : TStringList;
    FLastFilter          : TStringList;
    FFrameIndex          : Boolean;
    FShowGridMenu        : Boolean;
    FAutoSize            : Boolean;
    FAccessClpFormat     : Boolean;
    FFindFieldsOnPaste   : Boolean;
    FShowCheckBoxes      : Boolean;
    FNumMenus            : Integer;
    FExOptions           : ExOptionsSet;
    FOrigRC              : Integer;
    FExportDir           : String;
    FColWidthZero        : Integer;
    CurrentRowsHeigh     : Integer;
    GlobalOptions        : TDBGridOptions;
    CheckW               : Integer;
    CheckH               : Integer;
    CoordX               : Integer;
    CoordY               : Integer;
    ColumnsOrigWidth     : Array [0..300] of Integer;
    FUseBrackets         : Boolean;
    FBracketOpen: String;
    FBracketClose: String;
    Procedure            SetMenuTitles(Value:TStringList);
    Procedure            SetShowGridMenu(Value : Boolean);
    Procedure            SetShowCheckBoxes(Value : Boolean);
    Procedure            SetFrameIndex(Value : Boolean);
    Procedure            SetAutoSize(Value : Boolean);
    Procedure            SetExOptions(Value : ExOptionsSet);
  protected
    { Protected declarations }
    Procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    Procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    Procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    Procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    Procedure DrawColumnCell(const Rect: TRect; DataCol: Integer; Column: TColumn; State: TGridDrawState); override;
    Procedure DblClick; override;
    Procedure ColEnter;override;
    Function  IsInFieldList(FieldName, FieldList:String): Boolean;
    Procedure SizeColumns;
    Function  CalcTitlesHeight : Integer;
    Procedure RowHeightsChanged; override;
    Procedure LayoutChanged; override;
    Procedure SetColumnAttributes; override;
    Procedure Loaded; override;
    Procedure LinkActive(Value: Boolean);override;
    procedure SetUseBrackets(const Value: Boolean);
    procedure SetBracketClose(const Value: String);
    procedure SetBracketOpen(const Value: String);
    Procedure OnSelectIndex(Sender: TObject);
    Procedure OnSortBy(Sender: TObject);
    Procedure OnFilterBy(Sender: TObject);
    Procedure OnFind(Sender: TObject);
    Procedure OnSeek(Sender: TObject);
    Procedure OnQuickFind(Sender: TObject);
    Procedure OnQuickSeek(Sender: TObject);
    Procedure OnAutoSize(Sender: TObject);
    Procedure OnFrameIndexFields(Sender: TObject);
    Procedure OnChooseVisibleFields(Sender: TObject);
    Procedure OnSaveAs(Sender: TObject);
    Procedure OnSelectAll(Sender: TObject);
    Procedure OnDeSelectAll(Sender: TObject);
    Function  IsAccessDataInClipboard(Clp:TClipboard):Boolean;
    Function  StrToHex(S:String) : String;
    Function  HexToStr(S:String) : String;
    Function  GetFieldAsString(Field:TField):String;
    Function  GetFieldAsHEXString(Field:TField):String;
    Procedure CopyToClipboard;
    Function  PropertyExists(PropObject : Properties;PropertyName:String):Boolean;
    Function  GetCaption(FieldName : String):String;
    Function  FindField(FieldName:String; MSAccess:Boolean):TField;
    Function  FindColumn(FieldName:String):TColumn;
    Procedure ParseHexLine(RD,FN: TStringList; Line:String);
    Procedure ParseTextLine(RD : TStringList; Line:String);
    Function  ConvertToTS(S:String):TDateTime;
    Function  CountQuotes (S : String) : Integer;
    Function  CountTabs (S : String) : Integer;
    Function  ReplaceTabQuotes (Var S : String) : Integer;
    Function  ReplaceDoubleQuotes (Var S : String) : Integer;
    Procedure PreprocessRecords(Records : TStringList);
    Function  GetTextDateTime(S:String):TDateTime;
    Procedure PasteNativeFormat(CFFormat : Word; Clp:TClipboard);
    Procedure PasteTextFormat(CFFormat : Word; Clp:TClipboard);
    Procedure PasteFromClipboard;
    Procedure OnCutToClp(Sender: TObject);
    Procedure OnCopyToClp(Sender: TObject);
    Procedure OnPasteFromClp(Sender: TObject);
    Procedure OnUseAccessClpFormat(Sender: TObject);
    Procedure OnDelete(Sender: TObject);
    Procedure OnMailGridData(Sender: TObject);
    Function  StoreField(Table : TAdoTable; X:Integer): Boolean;
    Function  EscapeSpecialCharacters(Const S:String):String;
    Function  BracketField (const FieldName:String) : String;
    Procedure AccessExportToHTML(Table : TAdoTable; FileName:String; DeleteOld:Boolean);
    Procedure AccessExportToExcel(Table: TAdoTable; FileName, SheetName :String; ExcelVersion:Integer; DeleteOld:Boolean);
  public
    { Public declarations }
    Procedure   SelectAll;
    Procedure   DeSelectAll;
    Procedure   CutToClp;
    Procedure   CopyToClp;
    Procedure   PasteFromClp;
    Property    GridPopupMenu : TPopupMenu  Read  FPopupMenu;
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
  published
    { Published declarations }
    Property AutoSizeColumns             : Boolean      Read FAutoSize           Write SetAutoSize;
    Property ExOptions                   : ExOptionsSet Read FExOptions          Write SetExOptions;
    Property FindFieldsOnPaste           : Boolean      Read FFindFieldsOnPaste  Write FFindFieldsOnPaste;
    Property ShowGridMenu                : Boolean      Read FShowGridMenu       Write SetShowGridMenu;
    Property ShowCheckBoxes              : Boolean      Read FShowCheckBoxes     Write SetShowCheckBoxes;
    Property OfficeClipboardFormat       : Boolean      Read FAccessClpFormat    Write FAccessClpFormat;
    Property MenuTitles                  : TStringList  Read FMenuTitles         Write SetMenuTitles;
    Property FrameIndexFields            : Boolean      Read FFrameIndex         Write SetFrameIndex;
    Property DefaultExportDir            : String       Read FExportDir          Write FExportDir;
    Property UseBrackets                 : Boolean      read FUseBrackets        write SetUseBrackets;
    Property BracketOpen                 : String       read FBracketOpen        write SetBracketOpen;
    Property BracketClose                : String       read FBracketClose       write SetBracketClose;
  end;

Procedure Register;

implementation
{$R 'KA.Data.Ado.Controls.resx'}

Uses
  System.Text,
  System.Runtime.InteropServices,
  Borland.Vcl.Forms,
  Borland.Vcl.Dialogs,
  Borland.Vcl.Variants,
  KA.Data.Ado.HRollForm,
  KA.Data.Ado.SelectIndexUnit,
  KA.Data.Ado.FilterByUnit,
  KA.Data.Ado.SortByDialogUnit,
  KA.Data.Ado.FindSeekUnit,
  KA.Data.Ado.QuickFindSeekUnit,
  KA.Data.Ado.DateTime,
  KA.Data.Ado.CVCUnit,
  KA.Data.Ado.SASUnit;

Const
   ResourceBaseName  = 'KA.Data.Ado.Controls';
   dbOpenTable       = 1;
   dbOpenForwardOnly = 8;
   MAPI_DIALOG        = $00000008;
   MAPI_LOGON_UI      = $00000001;
   MAPI_NEW_SESSION   = $00000002;
   GUID_VALID_CHARS   = ['{','}','-','0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F','a','b','c','d','e','f'];

type
  FLAGS   = Cardinal;
  LHANDLE = Cardinal;

  [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
  MapiMessage = packed record
    ulReserved: Cardinal;         { Reserved for future use (M.B. 0)       }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszSubject: string;          { Message Subject                        }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszNoteText: string;         { Message Text                           }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszMessageType: string;      { Message Class                          }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszDateReceived: string;     { in YYYY/MM/DD HH:MM format             }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszConversationID: string;   { conversation thread ID                 }
    flFlags: FLAGS;               { unread,return receipt                  }
    lpOriginator: IntPtr;         { PMapiRecipDesc - Originator descriptor }
    nRecipCount: Cardinal;        { Number of recipients                   }
    lpRecips: IntPtr;             { PMapiRecipDesc - Recipient descriptors }
    nFileCount: Cardinal;         { # of file attachments                  }
    lpFiles: IntPtr;              { PMapiFileDesc - Attachment descriptors }
  end;
  TMapiMessage = MapiMessage;

  type
  [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
  MapiFileDesc = packed record
    ulReserved1 : Cardinal;        { Reserved for future use (must be 0)     }
    flFlags1    : Cardinal;           { Flags                                   }
    nPosition1  : Cardinal;         { character in text to be replaced by attachment }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszPathName1 : string;        { Full path name of attachment file       }
    lpszFileName1 : IntPtr;        { Original file name (optional)           }
    lpFileType1   : IntPtr;          { Pointer - Attachment file type (can be lpMapiFileTagExt) }

    ulReserved2 : Cardinal;        { Reserved for future use (must be 0)     }
    flFlags2    : Cardinal;           { Flags                                   }
    nPosition2  : Cardinal;         { character in text to be replaced by attachment }
    [MarshalAs(UnmanagedType.LPStr)]
    lpszPathName2 : string;        { Full path name of attachment file       }
    lpszFileName2 : IntPtr;        { Original file name (optional)           }
    lpFileType2   : IntPtr;          { Pointer - Attachment file type (can be lpMapiFileTagExt) }
  end;
  TMapiFileDesc = MapiFileDesc;

[DllImport('MAPI32.DLL', CharSet = CharSet.Ansi, SetLastError = True, EntryPoint = 'MAPISendMail')]
function MapiSendMail(lhSession: LHANDLE; ulUIParam: Cardinal; const lpMessage: TMapiMessage; flFlags: FLAGS; ulReserved: Cardinal): Cardinal; external;


function OleVarDataToVariant(OleVarData: TOleVarData): Variant;
var
  Buffer: IntPtr;
begin
  Buffer := Marshal.AllocHGlobal(Marshal.SizeOf(TypeOf(TOleVarData)));
  try
    Marshal.StructureToPtr(OleVarData, Buffer, False);
    Result := Variant(Marshal.GetObjectForNativeVariant(Buffer));
  finally
    Marshal.FreeHGlobal(Buffer);
  end;
end;


constructor TAdoDBGrid.Create(AOwner: TComponent);
Var
 MI  : TMenuItem;
 X   : Integer;
 BMP : TBitmap;
Begin
 Inherited Create(AOwner);
 FUseBrackets       := False;
 FBracketOpen       := '[';
 FBracketClose      := ']';

 FShowGridMenu      := True;
 FShowCheckBoxes    := True;
 FAccessClpFormat   := False;
 FFindFieldsOnPaste := True;
 FClicked           := False;
 FMenuBMP           := TBitmap.Create;
 FASCBMP            := TBitmap.Create;
 FDESCBMP           := TBitmap.Create;

 FMenuBMP.LoadFromResourceName('MENUBMP',ResourceBaseName,Assembly.GetExecutingAssembly);
 FASCBMP.LoadFromResourceName('ASC',ResourceBaseName,Assembly.GetExecutingAssembly);
 FDESCBMP.LoadFromResourceName('DESC',ResourceBaseName,Assembly.GetExecutingAssembly);

 FMenuBMP.Transparent      := True;
 FMenuBMP.TransparentColor := clSilver;

 FASCBMP.Transparent       := True;
 FASCBMP.TransparentColor  := clSilver;

 FDESCBMP.Transparent      := True;
 FDESCBMP.TransparentColor := clSilver;


 FMenuTitles        := TStringList.Create;
 FOriginalTitles    := TStringList.Create;
 FLastFilter        := TStringList.Create;
 FFrameIndex        := False;
 FAutoSize          := False;
 FOrigRC            := RowCount;
 CurrentRowsHeigh    := -1;
 GlobalOptions       := [];
 FExportDir         := '';

 {00} FMenuTitles.Add('Select Index'#9'Ctrl+I');
 {01} FMenuTitles.Add('Sort');
 {02} FMenuTitles.Add('Filter');
 {03} FMenuTitles.Add('-');
 {04} FMenuTitles.Add('Find');
 {05} FMenuTitles.Add('Seek');
 {06} FMenuTitles.Add('Quick Find');
 {07} FMenuTitles.Add('Quick Seek');
 {08} FMenuTitles.Add('-');
 {09} FMenuTitles.Add('Frame Index Fields'#9'Ctrl+R');
 {10} FMenuTitles.Add('Auto Size Columns'#9'Ctrl+Z');
 {11} FMenuTitles.Add('Choose Visible Columns'#9'Ctrl+O');
 {12} FMenuTitles.Add('-');
 {13} FMenuTitles.Add('Save as - export'#9'Ctrl+S');
 {14} FMenuTitles.Add('Mail as XLS/HTML'#9'Ctrl+T');
 {15} FMenuTitles.Add('-');
 {16} FMenuTitles.Add('Select All'#9'Ctrl+A');
 {17} FMenuTitles.Add('Deselect All'#9'Ctrl+N');
 {18} FMenuTitles.Add('Cut'#9'Ctrl+X');
 {19} FMenuTitles.Add('Copy'#9'Ctrl+C');
 {20} FMenuTitles.Add('Paste'#9'Ctrl+V');
 {21} FMenuTitles.Add('Office Clipboard'#9'Ctrl+B');

 FOriginalTitles.Assign(FMenuTitles);

 FNumMenus        := FMenuTitles.Count;
 FPopupMenu       := TPopupMenu.Create(Self);
 FPopupMenu.Name  := 'OptionsMenu';
 For X := 1 to FNumMenus do
     Begin
      MI := TMenuItem.Create(FPopupMenu);
      FPopupMenu.Items.Insert(0,MI);
     End;
 FPopupMenu.Items[0].OnClick  := OnSelectIndex;
 FPopupMenu.Items[1].OnClick  := OnSortBy;
 FPopupMenu.Items[2].OnClick  := OnFilterBy;
 FPopupMenu.Items[4].OnClick  := OnFind;
 FPopupMenu.Items[5].OnClick  := OnSeek;
 FPopupMenu.Items[6].OnClick  := OnQuickFind;
 FPopupMenu.Items[7].OnClick  := OnQuickSeek;

 FPopupMenu.Items[9].OnClick := OnFrameIndexFields;
 FPopupMenu.Items[10].OnClick := OnAutoSize;
 FPopupMenu.Items[11].OnClick := OnChooseVisibleFields;

 FPopupMenu.Items[13].OnClick := OnSaveAs;
 FPopupMenu.Items[14].OnClick := OnMailGridData;

 FPopupMenu.Items[16].OnClick := OnSelectAll;
 FPopupMenu.Items[17].OnClick := OnDeSelectAll;
 FPopupMenu.Items[18].OnClick := OnCutToClp;
 FPopupMenu.Items[19].OnClick := OnCopyToClp;
 FPopupMenu.Items[20].OnClick := OnPasteFromClp;
 FPopupMenu.Items[21].OnClick := OnUseAccessClpFormat;

 BMP        := TBitmap.Create;
 BMP.Handle := LoadBitmap(0, MAKEINTRESOURCE(32759));
 CheckW     := BMP.Width  div 4;
 CheckH     := BMP.Height div 3;
 BMP.Free;

 FColWidthZero := CheckW;
End;


destructor  TAdoDBGrid.Destroy;
Begin
 FDESCBMP.Free;
 FASCBMP.Free;
 FMenuBMP.Free;
 FPopupMenu.Free;
 FMenuTitles.Free;
 FOriginalTitles.Free;
 FLastFilter.Free;
 Inherited Destroy;
End;

Procedure TAdoDBGrid.SetExOptions(Value : ExOptionsSet);
Var
  RC : Integer;
  OO : ExOptionsSet;
Begin
  OO          := FExOptions;
  FExOptions := Value;
  if csLoading    in ComponentState Then Exit;
  if (NOT Assigned(DataSource)) Then Exit;
  if (NOT Assigned(DataSource.DataSet)) Then Exit;
  if (NOT DataSource.DataSet.Active) Then Exit;

  if csDesigning  in ComponentState Then
     Begin
       if  (exoSelectCheckBoxes in FExOptions)
       And (dgMultiSelect in Options) Then
           Begin
             SetColumnAttributes;
           End;

       if  (NOT (exoSelectCheckBoxes in FExOptions))
       And (exoSelectCheckBoxes in OO)
       And (dgMultiSelect in Options) Then
            Begin
             SetColumnAttributes;
            End;
       Exit;
     End;
  if exoWrapTitles in FExOptions Then
     Begin
       (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options + [goRowSizing];
       if dgTitles in Options Then
          Begin
           RowHeights[0] := CalcTitlesHeight;
           RecreateWnd;
          End;
       if NOT (exoRowSizing in FExOptions) Then
          (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options - [goRowSizing];

     End;
  if exoRowSizing in FExOptions Then
     Begin
       (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options + [goRowSizing];
     End
  Else
     Begin
       (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options - [goRowSizing];
       DefaultRowHeight := -1;
       Canvas.Font := Font;
       DefaultRowHeight := Canvas.TextHeight('I')+4;
       if dgTitles in Options Then
          Begin
           RowHeights[0] := CalcTitlesHeight;
          End;
       RecreateWnd;
     End;
  if exoLoadAllRecords in FExOptions Then
     Begin
        if  Assigned(DataSource)
        And Assigned(DataSource.Dataset)
        And (DataSource.Dataset is TAdoTable)
        And (DataSource.DataSet.Active) Then
            Begin
              RC := DataSource.DataSet.RecordCount;
              if dgTitles in Options Then Inc(RC);
              if RC > RowCount Then RowCount := RC;
            End;
     End
  Else
     Begin
       RowCount := FOrigRC;
     End;

  if (exoSelectCheckBoxes in FExOptions)
  And (dgMultiSelect in Options) Then SetColumnAttributes;

  if  (NOT (exoSelectCheckBoxes in FExOptions))
  And (exoSelectCheckBoxes in OO)
  And (dgMultiSelect in Options) Then SetColumnAttributes;
End;

Procedure TAdoDBGrid.Loaded;
Var
 X   : Integer;
 S   : String;
Begin
  Inherited Loaded;
  if FOriginalTitles.Count <> FMenuTitles.Count Then FMenuTitles.Assign(FOriginalTitles);
  For X := 0 To FMenuTitles.Count-1 do
      Begin
       if X < FPopupMenu.Items.Count Then
          Begin
            S:=FMenuTitles.Strings[X];
            if (Length(S) > 0) And (S[1]='!') Then Delete(S,1,1);
            FPopupMenu.Items[X].Caption:=S;
          End;
      End;
  SetAutoSize(FAutoSize);
End;

Procedure TAdoDBGrid.LinkActive(Value: Boolean);
Var
 X : Integer;
begin
    Inherited LinkActive(Value);

    if  Assigned(DataSource)
    And Assigned(DataSource.Dataset)
    And (DataSource.DataSet is TAdoTable)
    And (TAdoTable(DataSource.DataSet).TableDirect) Then
        Begin
         For X := 0 To Columns.Count-1 do
            Columns.Items[X].Width := Columns.Items[X].Width+FASCBMP.Width;
        End;
    if  (Assigned(DataSource))
    And (Assigned(DataSource.DataSet))
    And (DataSource.DataSet is TAdoTable)
    And (DataSource.DataSet.Active)
    And (DataSource.DataSet.State = dsBrowse) Then
        Begin
         FPopupMenu.Items[10].Checked := FAutoSize;
         FPopupMenu.Items[20].Checked := FAccessClpFormat;
         if (FExOptions <> []) Then ExOptions := FExOptions;
         if FAutoSize Then AutoSizeColumns:=True;
        End;
    //**************************************************** COMBO BOX BOOLEANS
    // If First field is ftBoolean then simulate ColEnter
    //****************************************************
    ColEnter;
    //**************************************************** COMBO BOX BOOLEANS
end;

Procedure TAdoDBGrid.SetMenuTitles(Value:TStringList);
Begin
 FMenuTitles.Assign(Value);
End;

Procedure TAdoDBGrid.SetShowGridMenu(Value : Boolean);
Begin
 FShowGridMenu:=Value;
 Refresh;
End;

Procedure TAdoDBGrid.SetShowCheckBoxes(Value : Boolean);
Var
 Tmp : Boolean;
Begin
 Tmp := FShowCheckBoxes;
 FShowCheckBoxes:=Value;
 If csLoading in ComponentState Then Exit;
 if Tmp Then
    Begin
      if (dgEditing in Options)
      And (NOT (goEditing in (Self AS TCustomGrid).Options)) Then
          (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options + [goEditing];
    End;
 ColEnter;                                                                         
 Refresh;
End;

Procedure TAdoDBGrid.SetFrameIndex(Value : Boolean);
Begin
 FFrameIndex := Value;
 FPopupMenu.Items[9].Checked := Value;
 Refresh;
End;

Procedure TAdoDBGrid.SetAutoSize(Value : Boolean);
Var
 X   : Integer;
 Vis : Boolean;
Begin
 FAutoSize := Value;
 if csLoading In ComponentState Then Exit;
 Vis := Visible;
 if FAutoSize Then
    Begin
      Hide;
      Try
       For X := 0 To Columns.Count-1 do
          Begin
            if Assigned(Columns.Items[X].Field) Then
               ColumnsOrigWidth[X] := Columns.Items[X].Width;
          End;
       SizeColumns;
      Finally
       if Vis then Show;
      End;
   End
 Else
   Begin
      Hide;
      Try
        For X := 0 To Columns.Count-1 do
          Begin
            if Assigned(Columns.Items[X].Field) Then
               Begin
                 if ColumnsOrigWidth[X] > 0 Then
                    Columns.Items[X].Width := ColumnsOrigWidth[X]
                 Else
                    if Columns.Items[X].Width > 0 Then
                       ColumnsOrigWidth[X] := Columns.Items[X].Width;
               End;
          End;
      Finally
       if Vis then Show;
      End;
   End;
 Visible := Vis;
End;

Function TAdoDBGrid.CalcTitlesHeight : Integer;
Var
 X     : Integer;
 H     : Integer;
 Title : String;
 ARect : Trect;
Begin
 Result := 0;
 For X := 0 to Columns.Count-1 do
     Begin
       Canvas.Font := Columns.Items[X].Title.Font;
       if (exoWrapTitles in FExOptions) Then
          Begin
            Title       := Columns.Items[X].Title.Caption;
            ARect       := CellRect(X,0);
            ARect.Left  := 0;
            ARect.Right := Columns.Items[X].Width-2;
            H := DrawText(Canvas.Handle, Title, Length(Title), ARect, DT_CALCRECT OR DT_WORDBREAK OR DT_EXPANDTABS OR DT_NOPREFIX);
          End
       Else
          Begin
           H := Canvas.TextHeight(Columns.Items[X].Title.Caption);
          End;
       if H > Result Then Result := H;
     End;
 Result := Result + 4;
End;

Procedure TAdoDBGrid.RowHeightsChanged;
Var
  RH : Integer;
  X  : Integer;
  BR : Integer;
Begin
  if  (exoRowSizing in FExOptions) Then
      Begin
       RH := DefaultRowHeight;
       BR := 0;
       If dgTitles in Options Then Inc(BR);
       For X := BR To RowCount-1 do
           Begin
             if RowHeights[X] <> RH Then
                Begin                                                       
                  Inherited DefaultRowHeight := RowHeights[X];
                  CurrentRowsHeigh :=  RowHeights[X];
                  if BR=1 Then RowHeights[0] := CalcTitlesHeight;
                  Break;
                End;
           End;
      End;
  Inherited RowHeightsChanged;
End;

Procedure TAdoDBGrid.LayoutChanged;
Begin
 inherited LayoutChanged;
 if (NOT Assigned(DataSource)) Then Exit;
 if (NOT Assigned(DataSource.DataSet)) Then Exit;
 if (NOT DataSource.DataSet.Active) Then Exit;
 if exoWrapTitles in FExOptions Then
     Begin
       (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options + [goRowSizing];
       if dgTitles in Options Then
          Begin
           RowHeights[0] := CalcTitlesHeight;
           RecreateWnd;
          End;
       if NOT (exoRowSizing in FExOptions) Then
          (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options - [goRowSizing];
     End;
 if  (exoRowSizing in FExOptions)
 And (CurrentRowsHeigh <> -1) Then
     Begin
       Inherited DefaultRowHeight := CurrentRowsHeigh;
       If dgTitles in Options Then RowHeights[0] := CalcTitlesHeight;
     End;
End;

Procedure TAdoDBGrid.SetColumnAttributes;
Begin
 Inherited SetColumnAttributes;

 if  (dgIndicator in Options)
 And (dgMultiSelect in Options)
 And (exoSelectCheckBoxes in ExOptions) Then
    Begin
      ColWidths[0] := CheckW*2;
    End;

 if  (dgIndicator in Options)
 And (dgMultiSelect in Options)
 And (NOT (exoSelectCheckBoxes in ExOptions)) Then
    Begin
      ColWidths[0] := FColWidthZero;
    End;
End;

Procedure TAdoDBGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell      : TGridCoord;
  SaveState : TGridState;
  Rect      : TRect;
  PT        : TPoint;
Begin
   SaveState := FGridState;
   PT.X := X;
   PT.Y := Y;
   if (SaveState = gsRowSizing) or (SaveState = gsColSizing) or (Not FShowGridMenu) or
      ((InplaceEditor <> nil) and (InplaceEditor.Visible) and
      (PtInRect(InplaceEditor.BoundsRect, PT))) then
       Begin
         inherited MouseDown(Button, Shift, X, Y);
         Exit;
       End;
 if  (Assigned(DataSource))
 And (Assigned(DataSource.DataSet))
 And (DataSource.DataSet.Active)
 And (DataSource.DataSet.State = dsBrowse) Then
     Begin
      Cell := MouseCoord(X,Y);
      if (Button = mbLeft) And (Cell.X =0 ) And (Cell.Y = 0) And (dgIndicator in Options) And (dgTitles in Options) then
         Begin
          //***********************************************************************
          // Upper Left Rectange is clicked - draw a flat style fixed cell
          //***********************************************************************
          FClicked := True;
          Rect := FTopRect;
          Canvas.Brush.Color := FixedColor;
          Canvas.FillRect(Rect);
          if (dgColLines in Options) Then
             Begin
              InflateRect(Rect, 1, 1);
              DrawEdge(Canvas.Handle, Rect, BDR_RAISEDINNER, BF_FLAT);
              DrawEdge(Canvas.Handle, Rect, BDR_RAISEDINNER, BF_FLAT);
             End;
          Canvas.Draw(((Rect.Left+Rect.Right-FMenuBMP.Width) div 2)+1,((Rect.Top + Rect.Bottom - FMenuBMP.Height) div 2)+1, FMenuBMP);
         End;
      If  (dgIndicator in Options)
      And (dgMultiSelect in Options)
      And (exoSelectCheckBoxes in ExOptions) Then
          Begin
            //***********************************************************************
            // CkeckMark
            //***********************************************************************
            if Cell.X = 0 Then
               Begin
                 if     (ssShift in Shift) Then Shift := Shift-[ssShift];
                 if     (ssAlt   in Shift) Then Shift := Shift-[ssAlt];
                 if Not (ssCtrl  in Shift) Then Shift := Shift+[ssCtrl];
               End
            Else
               Begin
               End;
           End;
     End;
 inherited MouseDown(Button, Shift, X, Y);
End;

Procedure TAdoDBGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell      : TGridCoord;
  SaveState : TGridState;
  Rect      : TRect;
  ShowMenu  : Boolean;
  MPT       : TPoint;
  PT        : TPoint;
begin
  CoordX    := X;
  CoordY    := Y;
  PT.X      := X;
  PT.Y      := Y;

  ShowMenu  := False;
  SaveState := FGridState;
  if (SaveState = gsRowSizing) or (SaveState = gsColSizing) or (Not FShowGridMenu) or
     ((InplaceEditor <> nil) and (InplaceEditor.Visible) and
     (PtInRect(InplaceEditor.BoundsRect, PT))) then
     Begin
       inherited MouseUp(Button, Shift, X, Y);
       Exit;
     End;
  if  (Assigned(DataSource))
  And (Assigned(DataSource.DataSet))
  And (DataSource.DataSet.Active)
  And (DataSource.DataSet.State = dsBrowse) Then
      Begin
       Cell := MouseCoord(X,Y);
       if ((Button = mbLeft) and (Cell.X =0 ) and (Cell.Y = 0) And (dgIndicator in Options) And (dgTitles in Options)) Or (FClicked) then
          Begin
           //***********************************************************************
           // Upper Left Rectange is clicked -
           // draw a raised style fixed cell and popup a context menu
           //***********************************************************************
           Rect := FTopRect;
           Canvas.Brush.Color := FixedColor;
           Canvas.FillRect(Rect);
           if (dgColLines in Options) Then
              Begin
                DrawEdge(Canvas.Handle, Rect, BDR_RAISEDINNER, BF_BOTTOMRIGHT);
                DrawEdge(Canvas.Handle, Rect, BDR_RAISEDINNER, BF_TOPLEFT);
              End;
           FClicked := False;
           ShowMenu  := True;
           //***********************************************************************
           Button:=mbRight;
          End;
      IF  (dgIndicator in Options)
      And (dgMultiSelect in Options)
      And (exoSelectCheckBoxes in ExOptions) Then
          Begin
           //***********************************************************************
           // CkeckMark
           //***********************************************************************
           if Cell.X = 0 Then
              Begin
                if     (ssShift in Shift) Then Shift := Shift-[ssShift];
                if     (ssAlt   in Shift) Then Shift := Shift-[ssAlt];
                if Not (ssCtrl  in Shift) Then Shift := Shift+[ssCtrl];
              End
           Else
              Begin
              End;
          End;
      End;
  inherited MouseUp(Button, Shift, X, Y);
  if  (Assigned(DataSource))
  And (Assigned(DataSource.DataSet))
  And (DataSource.DataSet.Active)
  And (DataSource.DataSet.State = dsBrowse)
  And (ShowMenu)  Then
     Begin
       MPT.X := FTopRect.Left;
       MPT.Y := FTopRect.Bottom;
       MPT := ClientToScreen(MPT);
       DrawCell(0,0,FTopRect,[gdFixed]);
       If  (DataSource.DataSet is TAdoTable)
       And (DataSource.DataSet.Active)
       And (DataSource.DataSet.State=dsBrowse) Then
           Begin
             For X := 0 to FNumMenus-1 do FPopupMenu.Items[X].Enabled := True;
             if TAdoTable(DataSource.Dataset).TableDirect Then
                Begin
                  FPopupMenu.Items[1].Enabled := False;
                  FPopupMenu.Items[2].Enabled := False;
                  if TAdoTable(DataSource.Dataset).IndexFieldCount=0 Then
                     Begin
                       FPopupMenu.Items[5].Enabled  := False;
                       FPopupMenu.Items[7].Enabled  := False;
                       FPopupMenu.Items[9].Enabled := False;
                     End;
                End
             Else
                Begin
                  FPopupMenu.Items[0].Enabled  := False;
                  FPopupMenu.Items[5].Enabled  := False;
                  FPopupMenu.Items[7].Enabled  := False;
                  FPopupMenu.Items[9].Enabled := False;
                End;
             FPopupMenu.Items[10].Checked := FAutoSize;
             FPopupMenu.Items[21].Checked := FAccessClpFormat;
              //****************************************************************
              // No SelectAll, Cut, Copy, Paste and Delete on EMPTY Dataset
              //****************************************************************
              if Pos('Microsoft.Jet.OLEDB.',TAdoTable(DataSource.DataSet).Connection.Provider)=0 Then
                 Begin
                   FPopupMenu.Items[13].Enabled := False;
                   FPopupMenu.Items[14].Enabled := False;
                 End;
              If (DataSource.DataSet.IsEmpty) Then
                Begin
                  FPopupMenu.Items[13].Enabled := False;
                  FPopupMenu.Items[16].Enabled := False;
                  FPopupMenu.Items[17].Enabled := False;
                  FPopupMenu.Items[18].Enabled := False;
                  FPopupMenu.Items[19].Enabled := False;
                End;
              //****************************************************************
              // No Cut, Paste and Delete on ReadOnly Dataset
              //****************************************************************
              if  (NOT DataSource.DataSet.CanModify)
              Or  (ReadOnly) Then
                 Begin
                  FPopupMenu.Items[18].Enabled := False;
                  FPopupMenu.Items[20].Enabled := False;
                 End;
              //****************************************************************
              // Hided by the Programmer
              //****************************************************************
              For X := 0 to FNumMenus-1 do
                 Begin
                   if X < FMenuTitles.Count-1 Then
                      Begin
                        if FMenuTitles.Strings[X][1]='!' Then FPopupMenu.Items[X].Visible := False;
                      End;
                 End;
             FPopupMenu.Popup(MPT.X,MPT.Y);
           End;
     End;
End;

Function TAdoDBGrid.IsInFieldList(FieldName, FieldList:String): Boolean;
Var
 SL : TStringList;
 FL : String;
 X  : Integer;
 L  : Integer;
Begin
 Result := False;
 FL     := FieldList;
 L      := Length(FL);
 if L=0 Then Exit;
 For X := 1 To L Do If FL[X]=';' Then FL[X]:=','; 
 SL     := TStringList.Create;
 Try
  SL.CommaText := FL;
  if SL.IndexOf(FieldName) <> -1 Then Result  := True;
 Except
 End;
 SL.Free;
End;

Procedure TAdoDBGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
Var
 Title     : String;
 Al        : Integer;
 Col_      : Integer;
 Row_      : Integer;
 Style     : UINT;
 CheckRect : TRect;
 OldActive : Integer;

 Column    : TColumn;
 ADBMP     : TBitmap;
 ID        : TIndexDef;
Begin
 inherited DrawCell(ACol, ARow, ARect, AState);
 if (AState <> [gdFixed])                  Then Exit;
 if (NOT Assigned(DataSource))             Then Exit;
 if (NOT Assigned(DataSource.DataSet))     Then Exit;
 If (NOT DataSource.DataSet.Active)        Then Exit;
 Col_ := ACol;
 if dgIndicator in Options Then Dec(Col_);
 Row_ := ARow;
 if (dgTitles in Options) Then  Dec(Row_);
 if  (dgIndicator in Options)
 And (dgTitles in Options)
 And (ACol=0)
 And (ARow=0) Then
    Begin
      if (FShowGridMenu) And (DataSource.DataSet.State = dsBrowse) Then
          Begin
           FTopRect := ARect;
           BeginUpdate;
           Canvas.Lock;
           Canvas.Draw((ARect.Left+ARect.Right-FMenuBMP.Width) div 2,(ARect.Top + ARect.Bottom - FMenuBMP.Height) div 2, FMenuBMP);
           Canvas.UnLock;
           EndUpdate;
          End;
    End
 Else
    Begin
      if  (ARow=0)
      And (exoWrapTitles in ExOptions)
      And (dgTitles in Options) Then
          Begin
            Title := Columns.Items[Col_].Title.Caption;
            BeginUpdate;
            Canvas.Lock;
            if (dgColLines in Options) Then InflateRect(ARect, -1, -1);
            Canvas.FillRect(ARect);
            Canvas.Font := Columns.Items[Col_].Title.Font;
            Canvas.Brush.Color := Columns.Items[Col_].Title.Color;
            Al := DT_LEFT;
            Case Columns.Items[Col_].Title.Alignment of
                 taLeftJustify   : Al := DT_LEFT;
                 taRightJustify  : Al := DT_RIGHT;
                 taCenter        : Al := DT_CENTER;
            End;
            DrawText(Canvas.Handle, Title, Length(Title), ARect, DT_WORDBREAK OR DT_EXPANDTABS OR DT_NOPREFIX Or AL);
            if (dgColLines in Options) Then
               Begin
                InflateRect(ARect, 1, 1);
                DrawEdge(Canvas.Handle, ARect, BDR_RAISEDINNER, BF_BOTTOMRIGHT);
                DrawEdge(Canvas.Handle, ARect, BDR_RAISEDINNER, BF_TOPLEFT);
               End;
            Canvas.UnLock;
            EndUpdate;
          End;
      if  (ACol=0)
      And (Row_ >= 0)
      And (dgIndicator in Options)
      And (dgMultiSelect in Options)
      And (exoSelectCheckBoxes in ExOptions) Then
          Begin
           //***********************************************************************
           // Draw Ckeckbox and CheckMark if the row is selected
           //***********************************************************************
           BeginUpdate;
           Canvas.Lock;
           CheckRect := ARect;
           CheckRect.Left := ARect.Left + 1;
           CheckRect.Right := CheckRect.Left + CheckW;
           CheckRect.Top := ARect.Top + (ARect.Bottom - ARect.Top - CheckH) div 2;
           CheckRect.Bottom := CheckRect.Top + CheckH;
           Style := DFCS_BUTTONCHECK;
           OldActive := DataLink.ActiveRecord;
           Try
             DataLink.ActiveRecord := Row_;
             if SelectedRows.IndexOf(Datalink.DataSet.Bookmark) <> -1 Then Style := DFCS_CHECKED;
           Except
           End;
           DataLink.ActiveRecord := OldActive;
           DrawFrameControl(Canvas.Handle, CheckRect, DFC_BUTTON, Style);
           Canvas.UnLock;
           EndUpdate;
         End;
      if (dgTitles in Options)
      And (ARow=0) Then
      Begin
        if (dgIndicator in Options) And (ACol=0) Then
            Begin
             //***************************************** NOTHING
            End
        Else
            Begin
              if  (DataSource.DataSet is TAdoTable)
              And (TAdoTable(DataSource.DataSet).TableDirect)
              And (TAdoTable(DataSource.DataSet).IndexName<> '') Then
                  Begin
                   Column := Columns[Col_];
                   ID     := TAdoTable(DataSource.Dataset).IndexDefs.Find(TAdoTable(DataSource.DataSet).IndexName);
                   if (ID <> Nil)
                   And (Assigned(Column))
                   And (Column.FieldName <> '') Then
                       Begin
                         if IsInFieldList(Column.FieldName, ID.DescFields) Then
                            ADBMP := FDESCBMP
                         Else
                         if IsInFieldList(Column.FieldName, ID.Fields) Then
                            ADBMP := FASCBMP
                         Else
                            ADBMP := Nil;
                         if ADBMP <> Nil Then
                            Begin
                              if (dgColLines in Options) Then InflateRect(ARect, -4, -2);
                              BeginUpdate;
                              Canvas.Lock;
                              Canvas.Draw(ARect.Right-ADBMP.Width, ARect.Top, ADBMP);
                              Canvas.UnLock;
                              EndUpdate;
                              if (dgColLines in Options) Then InflateRect(ARect, 4, 2);
                            End;
                       End;
                  End;
            End;
      End;
    End;
End;

Procedure TAdoDBGrid.DrawColumnCell(const Rect: TRect; DataCol: Integer; Column: TColumn; State: TGridDrawState);
Var
  X         : Integer;
  CheckRect : TRect;
  Style     : UINT;
Begin
 Inherited DrawColumnCell(Rect,DataCol,Column,State);
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (Assigned(Column.Field)) Then
     Begin
       if (FFrameIndex) And (TAdoTable(DataSource.Dataset).IndexFieldCount > 0) Then
          Begin
            For X := 0 To  TAdoTable(DataSource.Dataset).IndexFieldCount-1 do
                Begin
                  if WideLowerCase(Column.Field.FieldName)= WideLowerCase(TAdoTable(DataSource.Dataset).IndexFields[X].FieldName) Then
                     Begin
                       if Column.FieldName <> '' then
                          Begin
                           BeginUpdate;
                           Canvas.Lock;
                           Canvas.Brush.Color:=clRed;
                           Canvas.FrameRect(Rect);
                           Canvas.UnLock;
                           EndUpdate;
                          End;
                     End;
                End;
          End;
       //**************************************************** COMBO BOX BOOLEANS
       if  (FShowCheckBoxes)
       And (Assigned(Column.Field))
       And (Column.Field.DataType=ftBoolean) Then
          Begin
            if Column.FieldName <> '' then
               Begin
                 BeginUpdate;
                 Canvas.Lock;
                 Canvas.FillRect(Rect);
                 CheckRect := Rect;
                 CheckRect.Left := Rect.Left + (Rect.Right - Rect.Left - CheckW) div 2;
                 CheckRect.Right := CheckRect.Left + CheckW;
                 CheckRect.Top := Rect.Top + (Rect.Bottom - Rect.Top - CheckH) div 2;
                 CheckRect.Bottom := CheckRect.Top + CheckH;
                 if Column.Field.AsBoolean Then
                    Style := DFCS_CHECKED
                 Else
                    Style := DFCS_BUTTONCHECK;
                 DrawFrameControl(Canvas.Handle, CheckRect, DFC_BUTTON, Style);
                 Canvas.UnLock;
                 EndUpdate;
               End;
          End;
     End;
End;

Procedure TAdoDBGrid.DblClick;
Var
  Column : TColumn;
  GC     : TGridCoord;
  PozX   : Integer;
  PozY   : Integer;
Begin
  //**************************************************** COMBO BOX BOOLEANS
  GC   := MouseCoord(CoordX,CoordY);
  PozX := -1;
  PozY := -1;
  If  (dgIndicator in Options) Then PozX := 0;
  If  (dgTitles in Options)    Then PozY := 0;
  if  (GC.X > PozX)
  And (GC.Y > PozY)
  And Assigned(DataSource)
  And Assigned(DataSource.Dataset)
  And (DataSource.DataSet is TAdoTable)
  And (DataSource.DataSet.Active) Then
      Begin
        Column := Columns.Items[SelectedIndex];
        if  (FShowCheckBoxes)
        And (Assigned(Column.Field))
        And (Column.Field.DataType = ftBoolean)
        And (NOT ReadOnly)
        And (NOT (dgRowSelect in Options))
        And ((dgEditing in Options))
        And (DataSource.Dataset.CanModify)
        And (Not Column.Field.ReadOnly) Then
             Begin
                If (CanEditModify) Then
                   Begin
                    BeginUpdate;
                    Try
                     Column.Field.AsBoolean := Not Column.Field.AsBoolean;
                    Except
                    End;
                    inherited DblClick;
                    EndUpdate;
                    Exit;
                   End;
             End;
      End;
  //**************************************************** COMBO BOX BOOLEANS
  inherited DblClick;
End;

Procedure TAdoDBGrid.ColEnter;
Var
  Col : TColumn;
Begin
  if  (csLoading in ComponentState) Then Exit;
  If (Not Assigned(DataSource))
  Or (Not Assigned(DataSource.Dataset))
  Or (Not (DataSource.DataSet is TAdoTable))
  Or (Not (DataSource.DataSet.Active)) Then Exit;


  Col := Columns.Items[SelectedIndex];
  If  (FShowCheckBoxes)
  And ((dgEditing in Options))
  And (NOT (dgRowSelect in Options))
  And (Assigned(Col.Field)) Then
      Begin
        //*************************************************** COMBO BOX BOOLEANS
        If (Col.Field.DataType=ftBoolean) Then
           Begin
              if (goEditing in (Self AS TCustomGrid).Options) Then
                Begin
                  BeginUpdate;
                  (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options - [goEditing];
                  EndUpdate;
                End;
              if (goAlwaysShowEditor in (Self AS TCustomGrid).Options) Then
                Begin
                  BeginUpdate;
                  HideEditor;
                  EndUpdate;
                End;
           End
        //*************************************************** COMBO BOX BOOLEANS
        Else
           Begin
             if NOT (goEditing in (Self AS TCustomGrid).Options) Then
                Begin
                  BeginUpdate;
                  (Self AS TCustomGrid).Options := (Self AS TCustomGrid).Options + [goEditing];
                  EndUpdate;
                End;
           End;
      End;
 Inherited ColEnter;
End;

Procedure TAdoDBGrid.KeyDown(var Key: Word; Shift: TShiftState);
Var
 PozX : Integer;
 PozY : Integer;
Begin
 Inherited KeyDown(Key, Shift);
 If (Not Assigned(DataSource))
 Or (Not Assigned(DataSource.Dataset))
 Or (Not (DataSource.DataSet is TAdoTable))
 Or (Not (DataSource.DataSet.Active)) Then Exit;
 if (Key=Ord('A')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[16].Visible Then Exit;
      if NOT FPopupMenu.Items[16].Enabled Then Exit;
      OnSelectAll(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('N')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[17].Visible Then Exit;
      if NOT FPopupMenu.Items[17].Enabled Then Exit;
      OnDeSelectAll(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('I')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[0].Visible Then Exit;
      if NOT FPopupMenu.Items[0].Enabled Then Exit;
      OnSelectIndex(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('R')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[9].Visible Then Exit;
      if NOT FPopupMenu.Items[9].Enabled Then Exit;
      OnFrameIndexFields(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('Z')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[10].Visible Then Exit;
      if NOT FPopupMenu.Items[10].Enabled Then Exit;
      OnAutoSize(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('O')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[11].Visible Then Exit;
      if NOT FPopupMenu.Items[11].Enabled Then Exit;
      OnChooseVisibleFields(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('S')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[13].Visible Then Exit;
      if NOT FPopupMenu.Items[13].Enabled Then Exit;
      OnSaveAs(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('T')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[14].Visible Then Exit;
      if NOT FPopupMenu.Items[14].Enabled Then Exit;
      OnMailGridData(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('X')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[18].Visible Then Exit;
      if NOT FPopupMenu.Items[18].Enabled Then Exit;
      OnCutToClp(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('C')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[19].Visible Then Exit;
      if NOT FPopupMenu.Items[19].Enabled Then Exit;
      OnCopyToClp(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('V')) And (ssCtrl in Shift) And (Datalink.DataSet.State=dsBrowse) Then
    Begin
      if NOT FPopupMenu.Items[20].Visible Then Exit;
      if NOT FPopupMenu.Items[20].Enabled Then Exit;
      OnPasteFromClp(FPopupMenu);
      SetFocus;
    End;
 if (Key=Ord('B')) And (ssCtrl in Shift) Then
    Begin
      OnUseAccessClpFormat(FPopupMenu);
      SetFocus;
    End;
 //**************************************************** COMBO BOX BOOLEANS
 if (Key = VK_SPACE) Or (Key = VK_RETURN) Or (Key = VK_F2) Then
    Begin
      if NOT (dgRowSelect in Options)
      And    ((dgEditing in Options)) Then
         Begin
           If  (FShowCheckBoxes)
           And (Columns.Items[SelectedIndex].Field.DataType = ftBoolean) Then
              Begin
                if  (DataLink.DataSet.CanModify)
                And (Not ReadOnly)
                And (Not Columns.Items[SelectedIndex].Field.ReadOnly) Then
                    Begin
                     Key := 0;
                     BeginUpdate;
                     PozX := SelectedIndex;
                     PozY := SelectedIndex;
                     If  (dgIndicator in Options) Then Inc(PozX);
                     If  (dgTitles in Options)    Then Inc(PozY);
                     CoordX := CellRect(PozX,PozY).Left;
                     CoordY := CellRect(PozX,PozY).Top;
                     DblClick;
                     EndUpdate;
                    End;
              End;
         End;
    End;
 //**************************************************** COMBO BOX BOOLEANS
End;


Procedure TAdoDBGrid.SizeColumns;
Var
  RS : TDataset;
  CW : TStringList;
  X  : Integer;
  W  : Integer;
  S  : String;
  BK : TBookmarkStr;
Begin
  if  Assigned(DataSource)
  And Assigned(DataSource.Dataset)
  And (DataSource.DataSet is TAdoTable)
  And (DataSource.DataSet.Active)
  And (NOT DataSource.DataSet.IsEmpty) Then
      Begin
        CW := TStringList.Create;
        For X := 0 To Columns.Count-1 do
            Begin
              W := Canvas.TextWidth(Columns.Items[X].Title.Caption+'   ');
              if  (TAdoTable(DataSource.DataSet).TableDirect) Then Inc(W, FASCBMP.Width);
              CW.AddObject(Columns.Items[X].Field.FieldName,TObject(W));
            End;
        Enabled := False;
        DataSource.DataSet.DisableControls;
        Try
          RS := DataSource.DataSet;
          BK := RS.Bookmark;
          RS.First;
          While Not RS.EOF Do
            Begin
              For X := 0 To Columns.Count-1 do
                  Begin
                    if Fields[X].IsBlob Then
                       S := Fields[X].DisplayText
                    Else
                       S := Fields[X].Text;
                    W := Canvas.TextWidth(S+'   ');
                    if W > Integer(CW.Objects[X]) Then CW.Objects[X] := TObject(W);
                  End;
              RS.Next;
            End;
          RS.First;
          Try
            RS.Bookmark := BK;
          Except
          End;  
          For X := 0 To Columns.Count-1 Do Columns.Items[X].Width := Integer(CW.Objects[X]);
        Finally
          CW.Free;
          Enabled := True;
          DataSource.DataSet.EnableControls;
        End;
      End;
End;

//******************************************************************************
Procedure TAdoDBGrid.OnSelectIndex(Sender: TObject);
Var
  NewIndex : String;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
    Begin
      if TAdoTable(DataSource.Dataset).TableDirect Then
         Begin
           NewIndex:=TAdoTable(DataSource.Dataset).IndexName;
           Application.CreateForm(TAdoSelectIndex, AdoSelectIndex);
           if AdoSelectIndex.Execute(TAdoTable(DataSource.Dataset), NewIndex) Then
              Begin
              End;
           AdoSelectIndex.Free;
           TAdoTable(DataSource.Dataset).IndexName := NewIndex;
         End;
    End;
End;

Procedure TAdoDBGrid.OnSortBy(Sender: TObject);
Var
 DT : TAdoTable;
 ST : String;
 SL : TSTringList;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
      DT := TAdoTable(DataSource.Dataset);
      SL := TSTringList.Create;
      Application.CreateForm(TAdoSortByDialog,AdoSortByDialog);
      ST := DT.Sort;
      SL.CommaText := DT.Sort;
      if AdoSortByDialog.Execute(DT.FieldDefList,SL,False) Then
         Begin
           ST := StringReplace(SL.CommaText,'"','',[rfReplaceAll]);
           DT.Sort := ST;
         End
      Else
         Begin
           DT.Sort := ST;
         End;
      SL.Free;
      AdoSortByDialog.Free;
     End;
End;

Procedure TAdoDBGrid.OnFilterBy(Sender: TObject);
Var
  OldFilter   : String;
  OldFiltered : Boolean;
  NewFilter   : String;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
      Application.CreateForm(TAdoFilterBy, AdoFilterBy);
      AdoFilterBy.SQ := '[';
      AdoFilterBy.EQ := ']';
      NewFilter:=TAdoTable(DataSource.Dataset).Filter;
      if AdoFilterBy.Execute(TAdoTable(DataSource.Dataset), NewFilter, FLastFilter) Then
         Begin
           if NewFilter  <> '' Then
              Begin
               OldFilter   := TAdoTable(DataSource.Dataset).Filter;
               OldFiltered := TAdoTable(DataSource.Dataset).Filtered;
               Try
                 TAdoTable(DataSource.Dataset).Filter   := NewFilter;
                 TAdoTable(DataSource.Dataset).Filtered := True;
               Except
                 TAdoTable(DataSource.Dataset).Filter   := OldFilter;
                 TAdoTable(DataSource.Dataset).Filtered := OldFiltered;
               End;
              End
           Else
              Begin
               TAdoTable(DataSource.Dataset).Filtered := False;
              End;
         End;
      AdoFilterBy.Free;
     End;
End;

Procedure TAdoDBGrid.OnFind(Sender: TObject);
Var
 BK : TBookmarkStr;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
       BK := TAdoTable(DataSource.Dataset).Bookmark;
       Application.CreateForm(TAdoFindSeekDialog, AdoFindSeekDialog);
       if AdoFindSeekDialog.Execute(TAdoTable(DataSource.DataSet), False) Then
          Begin
          End
       Else
          Begin
            Try
              TAdoTable(DataSource.Dataset).Bookmark:=BK;
            Except
            end;
          End;
       AdoFindSeekDialog.Free;
     End;
End;

Procedure TAdoDBGrid.OnSeek(Sender: TObject);
Var
 BK : TBookmarkStr;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (TAdoTable(DataSource.Dataset).IndexFieldCount > 0)  Then
     Begin
       BK := TAdoTable(DataSource.Dataset).Bookmark;
       Application.CreateForm(TAdoFindSeekDialog, AdoFindSeekDialog);
       if AdoFindSeekDialog.Execute(TAdoTable(DataSource.DataSet), True) Then
          Begin
          End
       Else
          Begin
            Try
              TAdoTable(DataSource.Dataset).Bookmark:=BK;
            Except
            End;  
          End;
       AdoFindSeekDialog.Free;
     End;
End;

Procedure TAdoDBGrid.OnQuickFind(Sender: TObject);
Var
 BK : TBookmarkStr;
 FN : String;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
       FN := Self.SelectedField.FieldName;
       BK := TAdoTable(DataSource.Dataset).Bookmark;
       Application.CreateForm(TQFS, QFS);
       if QFS.Execute(TAdoTable(DataSource.DataSet), False, FN) Then
          Begin
          End
       Else
          Begin
            Try
              TAdoTable(DataSource.Dataset).Bookmark:=BK;
            Except
            End;  
          End;
       QFS.Free;
     End;
End;

Procedure TAdoDBGrid.OnQuickSeek(Sender: TObject);
Var
 BK : TBookmarkStr;
 FN : String;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (TAdoTable(DataSource.Dataset).IndexFieldCount > 0)  Then
     Begin
       FN := Self.SelectedField.FieldName;
       BK := TAdoTable(DataSource.Dataset).Bookmark;
       Application.CreateForm(TQFS, QFS);
       if QFS.Execute(TAdoTable(DataSource.DataSet), True, FN) Then
          Begin
          End
       Else
          Begin
            Try
              TAdoTable(DataSource.Dataset).Bookmark:=BK;
            Except
            End;  
          End;
       QFS.Free;
     End;
End;

Procedure TAdoDBGrid.OnAutoSize(Sender: TObject);
Begin
 FAutoSize := NOT FAutoSize;
 SetAutoSize(FAutoSize);
 FPopupMenu.Items[10].Checked := FAutoSize;
End;

Procedure TAdoDBGrid.OnFrameIndexFields(Sender: TObject);
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
       FPopupMenu.Items[9].Checked := Not FPopupMenu.Items[14].Checked;
       FFrameIndex := FPopupMenu.Items[9].Checked;
       Self.Refresh;
     End;
End;

Procedure TAdoDBGrid.OnChooseVisibleFields(Sender: TObject);
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active) Then
     Begin
       Application.CreateForm(TCVC, CVC);
       if CVC.Execute(Self) Then
          Begin
          End;
       CVC.Free;
       Self.Refresh;
     End;
End;

Procedure TAdoDBGrid.OnSaveAs(Sender: TObject);
Var
 Visible_   : TStringList;
 Invisible_ : TStringList;
 Table      : TAdoTable;
 FC         : TColumn;
 FF         : TField;
 X          : Integer;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (NOT DataSource.DataSet.IsEmpty) Then
     Begin
       Application.CreateForm(TSAS, SAS);
       SAS.UseBrackets    := FUseBrackets;
       SAS.BracketOpen    := BracketOpen;
       SAS.BracketClose   := BracketClose;
       Table              := TAdoTable(DataSource.DataSet);
       Visible_           := TStringList.Create;
       Invisible_         := TStringList.Create;
       Table.DisableControls;
       For X := 0 To Table.FieldCount-1 do
           Begin
            if Table.Fields[X].Visible Then
               Begin
                 FC := FindColumn(Table.Fields[X].FieldName);
                 IF FC = Nil Then Visible_.Add(Table.Fields[X].FieldName)
                    Else
                      if NOT FC.Visible Then Visible_.Add(Table.Fields[X].FieldName);
               End
            Else
               Begin
                 FC := FindColumn(Table.Fields[X].FieldName);
                 IF FC <> Nil Then
                    if FC.Visible Then Invisible_.Add(Table.Fields[X].FieldName);
               End;
           End;
       For X := 0 To Visible_.Count-1 do
             Begin
               FF := Table.FindField(Visible_.Strings[X]);
               if FF <> Nil Then FF.Visible := False;
             End;
       For X := 0 To Invisible_.Count-1 do
             Begin
               FF := Table.FindField(Invisible_.Strings[X]);
               if FF <> Nil Then FF.Visible := True;
             End;
       Try
         SAS.ExportDir := FExportDir;
         if SAS.ExecuteFromGrid(DataSource.DataSet, Self) Then
            Begin
            End;
       Finally
         For X := 0 To Invisible_.Count-1 do
             Begin
               FF := Table.FindField(Invisible_.Strings[X]);
               if FF <> Nil Then FF.Visible := False;
             End;
         For X := 0 To Visible_.Count-1 do
             Begin
               FF := Table.FindField(Visible_.Strings[X]);
               if FF <> Nil Then FF.Visible := True;
             End;
         Table.EnableControls;
         Invisible_.Free;
         Visible_.Free;
         SAS.Free;
       End;
     End;
End;

Procedure TAdoDBGrid.OnDeSelectAll(Sender: TObject);
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (TAdoTable(DataSource.Dataset).CursorType <> ctOpenForwardOnly)
 And (NOT DataSource.DataSet.IsEmpty) Then
     Begin
       if Not (dgMultiSelect In Options) Then
          Begin
            GlobalOptions := [];
            Exit;
          End
       Else                                                
          Begin
           SelectedRows.Clear;
           Refresh;
           if GlobalOptions = [dgMultiSelect] Then
              Begin
                Options       := Options-[dgMultiSelect];
                GlobalOptions := [];
              End;
          End;
     End;
End;
Procedure TAdoDBGrid.OnSelectAll(Sender: TObject);
Var
  BK : TBookmarkStr;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (TAdoTable(DataSource.Dataset).CursorType <> ctOpenForwardOnly)
 And (NOT DataSource.DataSet.IsEmpty) Then
     Begin
       if Not (dgMultiSelect In Options) Then
          Begin
            Options       := Options+[dgMultiSelect];
            GlobalOptions := [dgMultiSelect];
          End;
       BK := DataSource.DataSet.Bookmark;
       SelectedRows.Clear;
       Try
        Enabled:=False;
        DataSource.DataSet.DisableControls;
        DataSource.DataSet.First;
        While NOT DataSource.DataSet.EOF do
         Begin
          SelectedRows.CurrentRowSelected:=True;
          DataSource.DataSet.Next;
         End;
        Try
         DataSource.DataSet.Bookmark := BK;
        Except
        End;
       Finally
        DataSource.DataSet.EnableControls;
        Enabled:=True;
       End;
     End;
End;

Function TAdoDBGrid.StrToHex(S:String) : String;
Begin
 Result := Convert.ToBase64String(Encoding.Unicode.GetBytes(S));
End;

Function StreamToHex(MS : TMemoryStream) : String;
Begin
  Result := Convert.ToBase64String(MS.Memory);
End;

Function HexToStream(S:String):TBytes;
Begin
  Result := Convert.FromBase64String(S);
End;

Function TAdoDBGrid.HexToStr(S:String) : String;
Begin
 Result := Encoding.Unicode.GetString(Convert.FromBase64String(S));
End;

Function TAdoDBGrid.GetFieldAsHEXString(Field:TField):String;
Var
  DT  : Double;
  MS  : TMemoryStream;
Begin
  Result := '';
  if (Field.DataType=ftDate)
  OR (Field.DataType=ftTime)
  OR (Field.DataType=ftDateTime) Then
     Begin
      DT      := Field.AsDateTime;
      Result  := FloatToStr(DT);
     End
  Else
  if Field.DataType=ftBlob Then
     Begin
       MS := TMemoryStream.Create;
       Try
         (Field As TBlobField).SaveToStream(MS);
         MS.Size := (Field As TBlobField).BlobSize;
         Result := StreamToHex(MS);
       Finally
         MS.Free;
       End;
     End
  Else
  if Field.DataType=ftMemo Then
     Begin
       Result := StrToHex(Field.AsString);
     End
  Else
  if Field.DataType=ftString Then
     Begin
       Result := StrToHex(Field.AsString);
     End
  Else
     Begin
      Result := Field.AsString;
     End;
End;

Function TAdoDBGrid.GetFieldAsString(Field:TField):String;
Var
 Date_     : TDate;
 Time_     : TTime;
 DateTime_ : TDateTime;
 S         : String;
 P         : Integer;
 HasQuote  : Boolean;
Begin
  Result := '';
  if Field.DataType=ftDate Then
     Begin
      Date_  := Field.AsDateTime;
      if FAccessClpFormat Then
         Result := FormatDateTime('mm"/"dd"/"yyyy',Date_)
      Else
         Result := DateToStr(Date_);
     End
  Else
  if Field.DataType=ftTime Then
     Begin
       Time_ := Field.AsDateTime;
       if FAccessClpFormat Then
          Result := FormatDateTime('hh":"nn":"ss',Time_)
       Else
          Result := TimeToStr(Time_);
     End
  Else
  if Field.DataType=ftDateTime Then
     Begin
      DateTime_ := Field.AsDateTime;
      if FAccessClpFormat Then
         Result := FormatDateTime('mm"/"dd"/"yyyy hh":"nn":"ss',DateTime_)
      Else
         Result := DateTimeToStr(DateTime_);
     End
  Else
  if Field.DataType=ftBlob Then
     Begin
       if (Field as TBlobField).IsNull Then
          Result := '(blob)'
       Else
          Result := '(BLOB)';
     End
  Else
  if  (Field.DataType=ftBytes)
  And (Field.ValidChars=GUID_VALID_CHARS) Then
     Begin
       Result := Field.DisplayText;
     End
  Else
  if (Field.DataType=ftMemo) Or (Field.DataType=ftString) Then
     Begin
       S := Field.AsString;
       if FAccessClpFormat Then
          Begin
           if (Pos(#13,S) > 0) Or (Pos(#10,S) > 0) Then
              Begin
              P := Pos('"',S);
              While P > 0 Do
                Begin
                 Result := Result+Copy(S,1,P)+'"';
                 Delete(S,1,P);
                 P := Pos('"',S);
                End;
              Result := '"'+Result+S+'"';
              End
           Else
              Begin
              HasQuote := False;
              P := Pos('"',S);
              While P > 0 Do
                Begin
                  HasQuote := True;
                  Result := Result+Copy(S,1,P)+'"';
                  Delete(S,1,P);
                  P := Pos('"',S);
                End;
              if HasQuote Then
                Result := '"'+Result+S+'"'
              Else
                Result := Result+S;
              End;
          End
       Else
          Begin
            P := Pos(#13,S);
            While P > 0 Do
              Begin
               S[P] := ' ';
               P    := Pos(#13,S);
              End;
            P := Pos(#10,S);
            While P > 0 Do
              Begin
               Delete(S,P,1);
               P    := Pos(#10,S);
              End;
            Result := S;
          End;
     End
  Else
     Begin
      Result := Field.AsString;
     End;
End;

{$WARNINGS OFF}
Procedure TAdoDBGrid.CopyToClipboard;
Var
  BK          : TBookmarkStr;
  X           : Integer;
  Y           : Integer;
  L           : Integer;
  HeadText    : String;
  FTText      : String;
  CLPText     : String;
  HEXText     : String;
  Clp         : TClipboard;
  Locale      : LCID;
  Mem         : HGLOBAL;
  HEXMem      : HGLOBAL;
  PMem        : IntPtr;
  VisRow      : Integer;
  BA          : Array of Byte;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (NOT DataSource.DataSet.IsEmpty) Then
     Begin
       VisRow := Row;
       if dgTitles in Options Then Dec(VisRow);
       DataSource.DataSet.MoveBy(-VisRow);
       BK     := DataSource.DataSet.Bookmark;
       Locale := GetThreadLocale;
       SetThreadLocale(LOCALE_USER_DEFAULT);
       Clp := TClipboard.Create;
       Try
        Enabled:=False;
        DataSource.DataSet.DisableControls;
        CLPText  := '';
        HeadText := '';
        FTText   := '';
        //***************************************************** Copy Field Names
        For X := 0 To Columns.Count-1 do
            Begin
              if Columns.Items[X].Field.FieldKind=fkData Then
                 Begin
                  HeadText := HeadText + Columns.Items[X].Field.FieldName;
                  if X < Columns.Count-1 Then HeadText := HeadText+#9;
                 End;
            End;
        L := Length(HeadText);
        if HeadText[L]=#9 Then Delete(HeadText,L,1);
        HeadText := HeadText+#13#10;
        //**************************************************
        //  If this is enabled you cannot paste in MS Access
        //  However Access exports Field Names
        //**************************************************
        //  ClpText  := HeadText;
        //**************************************************
        HEXText  := HeadText;
        //************************************************* Copy Field DataTypes
        For X := 0 To Columns.Count-1 do
            Begin
              if Columns.Items[X].Field.FieldKind=fkData Then
                 Begin
                  FTText := FTText + IntToStr(Integer(Columns.Items[X].Field.DataType));
                  if X < Columns.Count-1 Then FTText := FTText+#9;
                 End;
            End;
        L := Length(FTText);
        if FTText[L]=#9 Then Delete(FTText,L,1);
        FTText  := FTText+#13#10;
        HEXText := HEXText+FTText;
        //***************************************************** Copy Actual Data
        if SelectedRows.Count > 0 Then
           Begin
             For Y := 0 To SelectedRows.Count-1 Do
               Begin
                DataSource.DataSet.Bookmark:=SelectedRows.Items[Y];
                For X := 0 To Columns.Count-1 do
                  Begin
                    if Columns.Items[X].Field.FieldKind=fkData Then
                       Begin
                        CLPText := CLPText + GetFieldAsString(Columns.Items[X].Field);
                        if X < Columns.Count-1 Then CLPText := CLPText+#9;

                        HEXText := HEXText + GetFieldAsHexString(Columns.Items[X].Field);
                        if X < Columns.Count-1 Then HEXText := HEXText+#9;
                       End;
                  End;
                L := Length(CLPText);
                if CLPText[L]=#9 Then Delete(CLPText,L,1);
                CLPText := CLPText+#13#10;

                L := Length(HEXText);
                if HEXText[L]=#9 Then Delete(HEXText,L,1);
                HEXText := HEXText+#13#10;
               End;
           End
        Else
           Begin
             For X := 0 To Columns.Count-1 do
                 Begin
                   if Columns.Items[X].Field.FieldKind=fkData Then
                      Begin
                       CLPText := CLPText + GetFieldAsString(Columns.Items[X].Field);
                       if X < Columns.Count-1 Then CLPText := CLPText+#9;

                       HEXText := HEXText + GetFieldAsHexString(Columns.Items[X].Field);
                       if X < Columns.Count-1 Then HEXText := HEXText+#9;
                      End;
                 End;
             L := Length(CLPText);
             if CLPText[L]=#9 Then Delete(CLPText,L,1);
             CLPText := CLPText+#13#10;

             L := Length(HEXText);
             if HEXText[L]=#9 Then Delete(HEXText,L,1);
             HEXText := HEXText+#13#10;
           End;

        //**********************************************************************
        Clp.Open;
        Clp.Clear;
        //************************* CF_UNICODETEXT
        BA   := Encoding.Unicode.GetBytes(CLPText);
        L    := Length(BA);
        Mem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, L);
        PMem := GlobalLock(Mem);
        Marshal.Copy(BA,0,PMem,L);
        GlobalUnlock(Mem);
        Clp.SetAsHandle(CF_UNICODETEXT,Mem);
        //************************* CF_TEXT
        BA   := Encoding.Convert(Encoding.Unicode,Encoding.GetEncoding(GETACP),Encoding.Unicode.GetBytes(CLPText));
        L    := Length(BA);
        Mem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, L);
        PMem := GlobalLock(Mem);
        Marshal.Copy(BA,0,PMem,L);
        GlobalUnlock(Mem);
        Clp.SetAsHandle(CF_TEXT,Mem);
        //************************* CFHEX_RECORDSET
        L       := Length(HEXText);
        HEXMem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, L);
        PMem    := GlobalLock(HEXMem);
        Marshal.Copy(BytesOf(HEXText),0,PMem,L);
        GlobalUnlock(HEXMem);
        Clp.SetAsHandle(RegisterClipboardFormat('HEX Recordset'),HEXMem);
        Clp.Close;
        //**********************************************************************
        Try
         DataSource.DataSet.Bookmark := BK;
         DataSource.DataSet.MoveBy(VisRow);
        Except
        End;
       Finally
        DataSource.DataSet.EnableControls;
        Enabled:=True;
        SetThreadLocale(Locale);
        Clp.Free;
       End;
     End;
End;
{$WARNINGS ON}

Function TAdoDBGrid.PropertyExists(PropObject : Properties;PropertyName:String):Boolean;
Var
  X : Integer;
Begin
  Result := False;
  For X := 0 to PropObject.Count-1 do
      Begin
        if WideCompareText(PropObject.Item[X].Name,PropertyName)=0 Then
           Begin
             Result := True;
             Exit;
           End;
      End;
End;

Function TAdoDBGrid.GetCaption(FieldName : String):String;
Var
  RS : Recordset;
Begin
  Result := FieldName;
End;

Function  TAdoDBGrid.FindColumn(FieldName:String):TColumn;
Var
 X : Integer;
Begin
 Result := Nil;
 For X := 0 To Columns.Count-1 do
     Begin
       if Columns.Items[X].Field.FieldName=FieldName Then
          Begin
            Result := Columns.Items[X];
          End;
     End;
End;

Function TAdoDBGrid.FindField(FieldName:String; MSAccess:Boolean):TField;
Var
  X    : Integer;
  CN   : String;
Begin
  Result := Nil;
  if NOT FFindFieldsOnPaste Then Exit;
  if MSAccess Then
     Begin
      For X := 0 To Columns.Count-1 do
      Begin
        CN := GetCaption(Columns.Items[X].Field.FieldName);
        if WideCompareText(CN,FieldName)=0 Then
           Begin
             Result := Columns.Items[X].Field;
             Exit;
           End;
      End;
     End;
  For X := 0 To Columns.Count-1 do
      Begin
        if WideCompareText(Columns.Items[X].Field.FieldName,FieldName)=0 Then
           Begin
             Result := Columns.Items[X].Field;
             Exit;
           End;
      End;
End;

Procedure TAdoDBGrid.ParseHexLine(RD,FN: TStringList; Line:String);
Var
 S   : String;
 Rec : String;
 P   : Integer;
 BR  : Integer;
 FT  : TFieldType;
Begin
 BR := 0;
 S  := Line;
 Repeat
  P := Pos(#9,S);
  if P > 0 Then
   Begin
    Rec := Copy(S,1,P-1);
    Delete(S,1,P);
    if BR < RD.Count Then
       Begin
        FT := TFieldType(Integer(FN.Objects[BR]));
        if (FT = ftString) or (FT=ftMemo) Then Rec := HexToStr(Rec);
        RD.Strings[BR] := Rec;
       End;
    Inc(BR);
   End;
  Until P=0;
  if BR < RD.Count Then
     Begin
       FT := TFieldType(Integer(FN.Objects[BR]));
       if (FT = ftString) or (FT=ftMemo) Then S := HexToStr(S);
       RD.Strings[BR]:=S;
     End;
End;


Function TAdoDBGrid.ConvertToTS(S:String):TDateTime;
Begin
 Result    := StrToFloat(S);
End;

{$WARNINGS OFF}
Procedure TAdoDBGrid.PasteNativeFormat(CFFormat : Word; Clp:TClipboard);
Var
 Records     : TStringList;
 FieldNames  : TStringList;
 RecordData  : TStringList;
 Mem         : HGLOBAL;
 PMem        : IntPtr;
 Sz          : Integer;
 S           : String;
 FN          : String;
 FT          : Integer;
 P           : Integer;
 BR          : Integer;
 FF          : TField;
 SourceFT    : TFieldType;
 X           : Integer;
 Y           : Integer;
 NumErrors   : Integer;
 BA          : TBytes;
 PasteWarning:String;
 MS          : TMemoryStream;
Begin
  Records     := TStringList.Create;
  FieldNames  := TStringList.Create;
  RecordData  := TStringList.Create;
  Enabled     := False;
  NumErrors   := 0;
  Try
    Clp.Open;
    Mem := Clp.GetAsHandle(CFFormat);
    If Mem <> 0 Then
       Begin
        Sz   := GlobalSize(Mem);
        PMem := GlobalLock(Mem);
        SetLength(BA,Sz);
        Marshal.Copy(PMem,BA,0,Sz);
        GlobalUnlock(Mem);
        S := StringOf(BA);
        Records.Text := S;
        S := '';
       End;
    Clp.Close;
    //**************************************************************************
    if Records.Count > 2 Then
       Begin
         //**************************************** Get Field Names
         S := Records.Strings[0];
         Repeat
           P := Pos(#9,S);
           if P > 0 Then
              Begin
                FN := Copy(S,1,P-1);
                Delete(S,1,P);
                FieldNames.Add(FN);
              End;
         Until P=0;
         FieldNames.Add(S);
         //**************************************** Get Field Types
         S  := Records.Strings[1];
         BR := 0;
         Repeat
           P := Pos(#9,S);
           if P > 0 Then
              Begin
                FN := Copy(S,1,P-1);
                FT := StrToInt(FN);
                Delete(S,1,P);
                if BR < FieldNames.Count Then FieldNames.Objects[BR]:=TObject(FT);
                Inc(BR);
              End;
         Until P=0;
         FT := StrToInt(S);
         if BR < FieldNames.Count Then FieldNames.Objects[BR]:=TObject(FT);
         //**************************************** Removes first two rows
         Records.Delete(0);
         Records.Delete(0);
         //**************************************** Actualy Add data
         RecordData.Assign(FieldNames);
         PasteWarning := 'You are about to paste '+IntToStr(Records.Count)+' records!'+#13#10+
                         'There will be no undo for this operation!'+#13#10+
                         'Are You sure?';
         If Application.MessageBox(PasteWarning,'Warning',MB_IconExclamation or MB_YesNoCancel)=ID_Yes
         Then
         if Records.Count > 0 Then
            Begin
              For X := 0 To FieldNames.Count-1 do
                  Begin
                    FF := FindField(FieldNames.Strings[X], False);
                    // ********************  Field not found by Name
                    // ********************  so it is by ordinal
                    if FF=Nil Then
                       Begin
                         if X < Columns.Count Then FF := Columns.Items[X].Field;
                       End;
                    //********************* Only Data Fields can be pasted
                    if (FF <> Nil) And (FF.FieldKind <> fkData) Then FF := Nil;
                    if (FF <> Nil) And (FF.ReadOnly) Then FF := Nil;
                    //********************* Duplication is NOT allowed
                    if FF <> Nil Then
                       Begin
                         if RecordData.IndexOfObject(FF)=-1 Then
                            RecordData.Objects[X] := FF
                         Else
                            RecordData.Objects[X] := Nil;
                       End
                    Else
                       Begin
                         RecordData.Objects[X] := FF;
                       End;
                   End;
              For Y := 0 To Records.Count-1 do
                  Begin
                    ParseHexLine(RecordData,FieldNames, Records.Strings[Y]);
                    Try
                     DataSource.DataSet.Append;
                     For X := 0 To RecordData.Count-1 do
                        Begin
                          FF := TField(RecordData.Objects[X]);
                          if FF <> Nil Then
                             Begin
                               SourceFT := TFieldType(Integer(FieldNames.Objects[X]));
                               if FF.DataType=ftDate Then
                                  Begin
                                    if (SourceFT=ftDate) or (SourceFT=ftTime) or (SourceFT=ftDateTime) Then
                                       Begin
                                         FF.AsDateTime := ConvertToTS(RecordData.Strings[X]);
                                       End
                                    Else
                                       Begin
                                         FF.AsString:=RecordData.Strings[X];
                                       End;
                                  End
                               Else
                               if FF.DataType=ftTime Then
                                  Begin
                                    if (SourceFT=ftDate) or (SourceFT=ftTime) or (SourceFT=ftDateTime) Then
                                       Begin
                                         FF.AsDateTime := ConvertToTS(RecordData.Strings[X]);
                                       End
                                    Else
                                       Begin
                                         FF.AsString:=RecordData.Strings[X];
                                       End;
                                  End
                               Else
                               if FF.DataType=ftDateTime Then
                                  Begin
                                    if (SourceFT=ftDate) or (SourceFT=ftTime) or (SourceFT=ftDateTime) Then
                                       Begin
                                         FF.AsDateTime := ConvertToTS(RecordData.Strings[X]);
                                       End
                                    Else
                                       Begin
                                         FF.AsString:=RecordData.Strings[X];
                                       End;
                                  End
                               Else
                               if FF.DataType=ftBlob Then
                                  Begin
                                    MS := TMemoryStream.Create;
                                    Try
                                       BA := HexToStream(RecordData.Strings[X]);
                                       MS.Write(BA,0,Length(BA));
                                       MS.Position := 0;
                                       (FF AS TBlobField).LoadFromStream(MS);
                                       MS.Position := 0;
                                    Finally
                                      MS.Free;
                                    End;
                                  End
                               Else
                                  Begin
                                    FF.AsString := RecordData.Strings[X];
                                  End;
                             End;
                        End;
                      DataSource.DataSet.Post;
                    Except
                      On E : Exception do
                         Begin
                           DataSource.DataSet.Cancel;
                           Inc(NumErrors);
                         End;
                    End;
                  End;
            End;
       End;
  Finally
   RecordData.Free;
   FieldNames.Free;
   Records.Free;
   Enabled := True;
  End;
  if NumErrors > 0 Then
    Begin
      ShowMessage(IntToStr(NumErrors) + ' records was not pasted due to an error!');
    End;
End;
{$WARNINGS ON}

Function  TAdoDBGrid.IsAccessDataInClipboard(Clp:TClipboard):Boolean;
Var
 CFFormat : Integer;
Begin
 Result := False;
 CFFormat := RegisterClipboardFormat('Data Records v8');  //****** Access 97
 If Clp.HasFormat(CFFormat) Then Result := True;
 CFFormat := RegisterClipboardFormat('Data Records v9');  //****** Access 2000
 If Clp.HasFormat(CFFormat) Then Result := True;
 CFFormat := RegisterClipboardFormat('Data Records v10'); //****** Access 2002?
 If Clp.HasFormat(CFFormat) Then Result := True;
End;


Function TAdoDBGrid.CountQuotes (S : String) : Integer;
var
  P : Integer;
Begin
    Result := 0;
    P := Pos('"', S);
    While P > 0 do
      Begin
        Inc (Result);
        Delete(S,1,P);
        P := Pos('"', S);
      End;
End;

Function TAdoDBGrid.CountTabs (S : String) : Integer;
var
  P : Integer;
Begin
    Result := 0;
    P := Pos(#9, S);
    While P > 0 do
      Begin
        Inc (Result);
        Delete(S,1,P);
        P := Pos(#9, S);
      End;
End;

Function TAdoDBGrid.ReplaceDoubleQuotes (Var S : String) : Integer;
Var
  P : Integer;
Begin
    Result := 0;
    P := Pos('""', S);
    While P > 0 do
      Begin
        Inc (Result);
        Delete(S,P,1);
        P := Pos('""', S);
      end;
End;

Function TAdoDBGrid.ReplaceTabQuotes (Var S : String) : Integer;
Var
  P : Integer;
  L : Integer;
Begin
    Result := 0;
    P := Pos(#9'"', S);
    While P > 0 do
      Begin 
        Inc (Result);
        Delete(S,P+1,1);
        S[P]:=#0;
        P := Pos(#9'"', S);
      end;

    P := Pos('"'#9, S);
    While P > 0 do
      Begin
        Inc (Result);
        Delete(S,P,1);
        S[P]:=#0;
        P := Pos('"'#9, S);
      end;

    P := Pos(#0, S);
    While P > 0 do
      Begin
        S[P]:=#9;
        P := Pos(#0, S);
      end;

    L := Length(S);
    if S[L]='"' Then Delete(S,L,1);
End;

Procedure TAdoDBGrid.PreprocessRecords(Records : TStringList);
Var
 X     : Integer;
 Q     : Integer;
 Equal : Boolean;
 S     : String;
 Tmp   : TStringList;
 MSS   : Boolean;
Begin
  Tmp  := TStringList.Create;
  Tmp.Assign(Records);
  MSS := True;
  Repeat
    Equal := True;
    For X := 0 to Records.Count-1 do
        Begin
          Q := CountQuotes(Records.Strings[X]);
          if (Q > 0) And Odd(Q) Then
             Begin
               Try
                 Records.Strings[X] := Records.Strings[X]+#13#10+Records.Strings[X+1];
                 Records.Delete(X+1);
                 Equal := False;
                 Break;
               Except
                 MSS   := False;
                 Equal := True;
                 Break;
               End;
             End;
        End;
  Until Equal;
  if Not MSS Then Records.Assign(Tmp);
  Tmp.Free;

  For X := 0 to Records.Count-1 do
      Begin
       S := Records.Strings[X];
       ReplaceTabQuotes(S);
       ReplaceDoubleQuotes(S);
       Records.Strings[X] := S;
      End;
End;

Procedure TAdoDBGrid.ParseTextLine(RD : TStringList; Line:String);
Var
 S   : String;
 Rec : String;
 P   : Integer;
Begin
 RD.Clear;
 S  := Line;
 Repeat
  P := Pos(#9,S);
  if P > 0 Then
   Begin
    Rec := Copy(S,1,P-1);
    Delete(S,1,P);
    RD.Add(Rec);
   End;
  Until P=0;
  RD.Add(S);
End;

Function TAdoDBGrid.GetTextDateTime(S:String):TDateTime;
Var
  DT : TAdoDateTime;
Begin
  Result := 0;
  DT := TAdoDateTime.Create(Self);
  DT.AccessShorting := False;
  Try
   DT.DateTimeString:=S;
   Result := DT.DateTime;
  Except
   DT.AccessShorting := True;
   Try
    DT.DateTimeString:=S;
    Result := DT.DateTime;
   Except
   End;
  End;
  DT.Free;
End;

{$WARNINGS OFF}
Procedure TAdoDBGrid.PasteTextFormat(CFFormat : Word; Clp:TClipboard);
Var
 MSAccess    : Boolean;
 Records     : TStringList;
 FieldNames  : TStringList;
 RecordData  : TStringList;
 Mem         : HGLOBAL;
 PMem        : IntPtr;
 Sz          : Integer;
 S           : String;
 FN          : String;
 P           : Integer;
 X           : Integer;
 Y           : Integer;
 NumFields   : Integer;
 FF          : TField;
 NumErrors   : Integer;
 DT          : TDateTime;
 PasteWarning: String;
 BA          : TBytes;
Begin
 MSAccess    := IsAccessDataInClipboard(Clp);
 Records     := TStringList.Create;
 FieldNames  := TStringList.Create;
 RecordData  := TStringList.Create;
 Enabled     := False;
 NumErrors   := 0;
 Try
    Clp.Open;
    Mem := Clp.GetAsHandle(CFFormat);
    If Mem <> 0 Then
       Begin
        Sz   := GlobalSize(Mem);
        PMem := GlobalLock(Mem);
        SetLength(BA,Sz);
        Marshal.Copy(PMem,BA,0,Sz);
        GlobalUnlock(Mem);
        S := StringOf(BA);
        GlobalUnlock(Mem);
        Records.Text:=S;
        S:='';
       End;
    Clp.Close;
    if Records.Count > 0 Then
       Begin
        if MSAccess Then
           Begin
             //**************************************** Get Field Names
             S := Records.Strings[0];
             Repeat
               P := Pos(#9,S);
               if P > 0 Then
                 Begin
                  FN := Copy(S,1,P-1);
                  Delete(S,1,P);
                  FieldNames.Add(FN);
                 End;
             Until P=0;
             FieldNames.Add(S);
             Records.Delete(0);
             For X := 0 To FieldNames.Count-1 Do
                 Begin
                   FF := FindField(FieldNames.Strings[X],True);
                   FieldNames.Objects[X]:= FF;
                 End;
           End;
        PreprocessRecords(Records);
        PasteWarning := 'You are about to paste '+IntToStr(Records.Count)+' records!'+#13#10+
                         'There will be no undo for this operation!'+#13#10+
                         'Are You sure?';
        If Application.MessageBox(PasteWarning,'Warning',MB_IconExclamation or MB_YesNoCancel)=ID_Yes
        Then
        if Records.Count > 0 Then
          Begin
           For Y := 0 To Records.Count-1 do
            Begin
              Try
               DataSource.DataSet.Append;
               NumFields := CountTabs(Records.Strings[Y])+1;
               ParseTextLine(RecordData,Records.Strings[Y]);
               //***************************************************************
               if MSAccess Then
                  Begin
                    For X := 0 To FieldNames.Count-1 Do
                        Begin
                          FF := TField(FieldNames.Objects[X]);
                          if (FF = Nil) and (X < Columns.Count) Then
                             Begin
                               FF := Columns.Items[X].Field;
                               if FF.FieldKind <> fkData Then FF := Nil;
                               if FF.ReadOnly Then FF := Nil;
                             End;
                          if FF <> Nil Then
                             Begin
                               if (FF.DataType=ftDate)
                               or (FF.DataType=ftTime)
                               or (FF.DataType=ftDateTime) Then
                                  Begin
                                     DT := GetTextDateTime(RecordData.Strings[X]);
                                     if DT <> 0 Then FF.AsDateTime := DT;
                                  End
                               Else
                                  Begin
                                    FF.AsString := RecordData.Strings[X];
                                  End;
                             End;
                        End;
                  End
               Else
                  Begin
                    For X := 0 To NumFields-1 Do
                        Begin
                          if X < Columns.Count Then
                             Begin
                              FF := Columns.Items[X].Field;
                              if  (NOT FF.ReadOnly)
                              And (FF.FieldKind = fkData) Then
                                  Begin
                                    if (FF.DataType=ftDate)
                                    or (FF.DataType=ftTime)
                                    or (FF.DataType=ftDateTime) Then
                                       Begin
                                         DT := GetTextDateTime(RecordData.Strings[X]);
                                         if DT <> 0 Then FF.AsDateTime := DT;
                                       End
                                    Else
                                       Begin
                                         FF.AsString := RecordData.Strings[X];
                                       End;
                                  End;
                             End;
                        End;
                  End;
               DataSource.DataSet.Post;
              Except
               Inc(NumErrors);
               DataSource.DataSet.Cancel;
              End;
            End;
          End;
       End;
 Finally
   RecordData.Free;
   FieldNames.Free;
   Records.Free;
   Enabled := True;
 End;
 if NumErrors > 0 Then
    Begin
      ShowMessage(IntToStr(NumErrors) + ' records was not pasted due to an error!');
    End;
End;
{$WARNINGS ON}

Procedure TAdoDBGrid.PasteFromClipboard;
Var
  Clp       : TClipboard;
  CFFormat  : Word;
Begin
 if  Assigned(DataSource)
 And Assigned(DataSource.Dataset)
 And (DataSource.DataSet is TAdoTable)
 And (DataSource.DataSet.Active)
 And (NOT ReadOnly) Then
  Begin
    Clp := TClipboard.Create;
    Try
     CFFormat := RegisterClipboardFormat('HEX Recordset');
     If Clp.HasFormat(CFFormat)  Then PasteNativeFormat(CFFormat, Clp)
        Else
          If Clp.HasFormat(CF_UNICODETEXT) Then PasteTextFormat(CF_UNICODETEXT, Clp)
             Else
               If Clp.HasFormat(CF_TEXT) Then PasteTextFormat(CF_TEXT, Clp);
    Finally
     Clp.Free;
    End;
  End;                       
End;

Procedure TAdoDBGrid.OnCutToClp(Sender: TObject);
Var
  PasteWarning : String;
  SelCount     : Integer;
Begin
  If ReadOnly Then Exit;
  SelCount := SelectedRows.Count;
  if SelCount=0 Then SelCount := 1;
  PasteWarning := 'You are about to cut '+IntToStr(SelCount)+' records!'+#13#10+
                   'There will be no undo for this operation!'+#13#10+
                   'Are You sure?';
  If Application.MessageBox(PasteWarning,'Warning',MB_IconExclamation or MB_YesNoCancel)=ID_Yes Then
     Begin
      CopyToClipboard;
      OnDelete(Sender);
     End;
End;

Procedure TAdoDBGrid.OnCopyToClp(Sender: TObject);
Begin
  CopyToClipboard;
End;

Procedure TAdoDBGrid.OnPasteFromClp(Sender: TObject);
Begin
 PasteFromClipboard;
End;

Procedure TAdoDBGrid.OnUseAccessClpFormat(Sender: TObject);
Begin
  FAccessClpFormat := Not FAccessClpFormat;
End;

Procedure TAdoDBGrid.OnDelete(Sender: TObject);
Var
 Key   : Word;
 Shift : TShiftState;
Begin
 Key    := VK_DELETE;
 Shift  := [ssCtrl];
 KeyDown(Key, Shift);
End;


Procedure TAdoDBGrid.SelectAll;
Begin
 OnSelectAll(Self);
End;

Procedure TAdoDBGrid.DeSelectAll;
Begin
  OnDeSelectAll(Self);
End;

Procedure TAdoDBGrid.CutToClp;
Begin
 OnCutToClp(Self);
End;

Procedure TAdoDBGrid.CopyToClp;
Begin
 OnCopyToClp(Self);
End;

Procedure TAdoDBGrid.PasteFromClp;
Begin
 OnPasteFromClp(Self);
End;

Function TAdoDBGrid.StoreField(Table : TAdoTable; X:Integer): Boolean;
Begin
   Case Table.Fields[X].FieldKind of
        fkData       : Result := True;
        fkCalculated : Result := False;
        fkLookup     : Result := False;
   Else                Result := False;
   End;
   if NOT Table.Fields[X].Visible Then Result := False;
End;

Function TAdoDBGrid.EscapeSpecialCharacters(Const S:String):String;
Var
 X, L : Integer;
Begin
 Result := '';
 L      := Length(S);
 if L    = 0 Then Exit;
 For X := 1 To L do
     Begin
       if (S[X]='"')  Then Result := Result+'"';
       if (S[X]='''') Then Result := Result+'''';
       if (S[X]='*')
       Or (S[X]='?')
       Or (S[X]='#')
       Or (S[X]='[')
       Or (S[X]=']')
       Or (S[X]='!')
       Or (S[X]='-')  Then Result := Result+'[';
       Result := Result+S[X];
       if (S[X]='*')
       Or (S[X]='?')
       Or (S[X]='#')
       Or (S[X]='[')
       Or (S[X]=']')
       Or (S[X]='!')
       Or (S[X]='-')  Then Result := Result+']';
     End;
End;


Function TAdoDBGrid.BracketField (const FieldName:String) : String;
Var
  P : Integer;
  S : String;
Begin
  Result := '';
  if UseBrackets Then
     Begin
       S := FieldName;
       P := Pos('.',S);
       While P > 0 Do
         Begin
           Result := Result+EscapeSpecialCharacters(Borland.Delphi.System.Copy(S,1,P-1));
           Result := BracketClose+Result+'.'+BracketOpen;
           Borland.Delphi.System.Delete(S,1,P);
           P := Pos('.',S);
         End;
       Result := BracketOpen+Result+EscapeSpecialCharacters(S)+BracketClose;
     End
  Else
     Begin
       Result := FieldName;
     End;
End;


Procedure TAdoDBGrid.AccessExportToHTML(Table : TAdoTable; FileName:String; DeleteOld:Boolean);
Var
 SQL : TStringList;
 FN  : String;
 FP  : String;
 SS  : String;
 X   : Integer;
 FC  : Integer;
 TN  : String;
 RA  : OleVariant;
Begin
 If Not Table.Active Then DatabaseError('Table must be open in order to export data!');
 TN := '';
 Try
 if (DeleteOld) And FileExists(FileName) Then DeleteFile(FileName);
 TN := Table.TableName;
 FN := ExtractFileName(FileName);
 FP := ExtractFilePath(FileName);
 if FP='' Then FP:='.';
 SS:='';
 FC := Table.FieldCount-1;
 For X := 0 To FC do
     Begin
      if StoreField(Table,X) Then
       Begin
        if UseBrackets Then
          Begin
           SS := SS+' '+BracketField(Table.Fields[X].FieldName);
           if X < FC Then SS := SS+','
          End
        Else
          Begin
           if X < FC Then
              SS := SS+' '+Table.Fields[X].FieldName+','
           Else
              SS := SS+' '+Table.Fields[X].FieldName;
          End;
       End;
     End;
 if SS[Length(SS)]=',' Then Borland.Delphi.System.Delete(SS,Length(SS),1);
 SQL := TStringList.Create;
 Try
   SQL.Add(Format('Select %s INTO [%s] IN "%s"[HTML Export;] FROM [%s]',[SS, FN,FP,TN]));
   if (Table.Filtered) And (Table.Filter <> '') Then SQL.Add(' WHERE '+Table.Filter);
   SQL.Add(';');
   Table.Connection.ConnectionObject.Execute(SQL.Text,RA,0)
 Finally
   SQL.Free;
 End;
 Finally
 End;
End;

Procedure TAdoDBGrid.AccessExportToExcel(Table: TAdoTable; FileName, SheetName :String; ExcelVersion:Integer; DeleteOld:Boolean);
Var
 SQL : TStringList;
 EV  : String;
 SS  : String;
 X   : Integer;
 FC  : Integer;
 TN  : String;
 RA  : OleVariant;
Begin
 If Not Table.Active Then DatabaseError('Table must be open in order to export data!');
 TN := '';
 Try
 if (DeleteOld) And FileExists(FileName) Then DeleteFile(FileName);
 TN := Table.TableName;
 EV := 'Excel 8.0;';
 Case ExcelVersion of
      3 : EV := 'Excel 3.0;';
      4 : EV := 'Excel 4.0;';
      5 : EV := 'Excel 5.0;';
      8 : EV := 'Excel 8.0;';
 End;
 SS:='';
 FC := Table.FieldCount-1;
 For X := 0 To FC do
     Begin
       if (Table.Fields[X].DataType<>ftBlob) And (StoreField(Table,X)) Then
          Begin
            if UseBrackets Then
               Begin
                SS := SS+' '+BracketField(Table.Fields[X].FieldName);
                if X < FC Then SS := SS+','
               End
            Else
               Begin
                if X < FC Then
                   SS := SS+' '+Table.Fields[X].FieldName+','
                Else
                   SS := SS+' '+Table.Fields[X].FieldName;
               End;
          End;
     End;
 if SS[Length(SS)]=',' Then Borland.Delphi.System.Delete(SS,Length(SS),1);
 SQL := TStringList.Create;
 Try
   SQL.Add(Format('Select %s INTO [%s] IN "%s"[%s] FROM [%s]',[SS,SheetName,FileName,EV,TN]));
   if (Table.Filtered) And (Table.Filter <> '') Then SQL.Add(' WHERE '+Table.Filter);
   SQL.Add(';');
   Table.Connection.ConnectionObject.Execute(SQL.Text,RA,0)
 Finally
   SQL.Free;
 End;
 Finally
 End;
End;


Procedure   TAdoDBGrid.OnMailGridData(Sender: TObject);
Const
  MAPIDLL  : String = 'MAPI32.DLL';
Var
  MSG        : TMapiMessage;
  FIL        : TMapiFileDesc;
  Err        : Integer;
  XLSFileName: String;
  HTMFileName: String;
  Path       : StringBuilder;
  TN         : String;
  Visible_   : TStringList;
  Invisible_ : TStringList;
  Table      : TAdoTable;
  FC         : TColumn;
  FF         : TField;
  X          : Integer;
  PFIL       : IntPtr;
Begin
  if  Assigned(DataSource)
  And Assigned(DataSource.Dataset)
  And (DataSource.DataSet is TAdoTable)
  And (DataSource.DataSet.Active)
  And (NOT DataSource.DataSet.IsEmpty) Then
     Begin
     //******************************************************* CREATE SNAPSHOT
     Table              := TAdoTable(DataSource.DataSet);
     Visible_           := TStringList.Create;
     Invisible_         := TStringList.Create;
     Table.DisableControls;
     For X := 0 To Table.FieldCount-1 do
           Begin
            if Table.Fields[X].Visible Then
               Begin
                 FC := FindColumn(Table.Fields[X].FieldName);
                 IF FC = Nil Then Visible_.Add(Table.Fields[X].FieldName)
                    Else
                      if NOT FC.Visible Then Visible_.Add(Table.Fields[X].FieldName);
               End
            Else
               Begin
                 FC := FindColumn(Table.Fields[X].FieldName);
                 IF FC <> Nil Then
                    if FC.Visible Then Invisible_.Add(Table.Fields[X].FieldName);
               End;
           End;
     For X := 0 To Visible_.Count-1 do
             Begin
               FF := Table.FindField(Visible_.Strings[X]);
               if FF <> Nil Then FF.Visible := False;
             End;
     For X := 0 To Invisible_.Count-1 do
             Begin
               FF := Table.FindField(Invisible_.Strings[X]);
               if FF <> Nil Then FF.Visible := True;
             End;
       //*********************************************************************
     Path := StringBuilder.Create(1000);
     Try
       Try
         //************************************************* Export to temp file
         TN               := Table.TableName;
         if TN='' Then TN := 'Table';
         GetTempPath(1000,Path);

         XLSFileName := Path.ToString+TN+'.xls';
         DeleteFile(XLSFileName);
         AccessExportToExcel(Table, XLSFileName, TN, 8, True);

         HTMFileName := Path.ToString+TN+'.htm';
         DeleteFile(HTMFileName);
         AccessExportToHTML(Table, HTMFileName, True);
         //*********************************************************************
       Finally
         //**************************************************** RESTORE SNAPSHOT
         For X := 0 To Invisible_.Count-1 do
             Begin
               FF := Table.FindField(Invisible_.Strings[X]);
               if FF <> Nil Then FF.Visible := False;
             End;
         For X := 0 To Visible_.Count-1 do
             Begin
               FF := Table.FindField(Visible_.Strings[X]);
               if FF <> Nil Then FF.Visible := True;
             End;
         Table.EnableControls;
         Invisible_.Free;
         Visible_.Free;
         //*********************************************************************
       End;
       //*************************************************** Use MAPI to send it
       FIL.ulReserved1         := 0;
       FIL.flFlags1            := 0;
       FIL.nPosition1          := Cardinal($FFFFFFFF);
       FIL.lpszPathName1       := XLSFileName;
       FIL.lpszFileName1       := Marshal.StringToHGlobalAnsi(TN+'.xls');
       FIL.lpFileType1         := Nil;

       FIL.ulReserved2         := 0;
       FIL.flFlags2            := 0;
       FIL.nPosition2          := Cardinal($FFFFFFFF);
       FIL.lpszPathName2       := HTMFileName;
       FIL.lpszFileName2       := Marshal.StringToHGlobalAnsi(TN+'.htm');
       FIL.lpFileType2         := Nil;

       PFIL := Marshal.AllocHGlobal(SizeOf(TMapiFileDesc));
       Try
         Marshal.StructureToPtr(FIL,PFIL,TRUE);

         MSG.ulReserved            := 0;
         MSG.lpszSubject           := TAdoTable(DataSource.Dataset).TableName;
         MSG.lpszNoteText          := Nil;
         MSG.lpszMessageType       := Nil;
         MSG.lpszDateReceived      := Nil;
         MSG.lpszConversationID    := Nil;
         MSG.flFlags               := 0;
         MSG.lpOriginator          := Nil;
         MSG.nRecipCount           := 0;
         MSG.nFileCount            := 2;
         MSG.lpFiles               := PFIL;
         Err                       := MapiSendMail(0, Application.Handle, MSG, MAPI_DIALOG or MAPI_LOGON_UI or MAPI_NEW_SESSION, 0);
         if Err > 1 Then ShowMessage(SysErrorMessage(ERR));
       Finally
         Marshal.FreeHGlobal(PFIL);
       End;
     Finally
       DeleteFile(HTMFileName);
       DeleteFile(XLSFileName);
       DeleteFile(Path.ToString+'schema.ini');
       Path.Free;
      End;
   End;
End;

procedure TAdoDBGrid.SetUseBrackets(const Value: Boolean);
begin
  FUseBrackets := Value;
end;

procedure TAdoDBGrid.SetBracketOpen(const Value: String);
begin
  FBracketOpen := Value;
end;

procedure TAdoDBGrid.SetBracketClose(const Value: String);
begin
  FBracketClose := Value;
end;

Procedure Register;
begin
  RegisterComponents('Ado Controls', [TAdoDBGrid]);
end;


end.
